// (C) 2017-2020 by folkert van heusden, released under AGPL v3.0
#include <ctype.h>
#include <poll.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "error.h"
#include "source.h"
#include "source_pixelflood.h"
#include "picio.h"
#include "filter.h"
#include "log.h"
#include "utils.h"

typedef struct {
	std::string id;
	uint8_t *fb;
	int fd;
	int width, height;
	int listen_port;
	std::atomic_bool *local_stop_flag;
} pf_s_parameter_t;

void * udp_receive_thread(void *p)
{
	// FIXME
	return nullptr;
}

int fromhex(int c)
{
	c = toupper(c);

	if (c >= 'A')
		return c - 'A' + 10;

	return c - '0';
}

bool handle_command(pf_s_parameter_t *pp, char *const line)
{
	if (strncmp(line, "SIZE", 4) == 0) {
		char out[32];
		int len = snprintf(out, sizeof out, "SIZE %d %d\n", pp -> width, pp -> height);

		return WRITE(pp -> fd, out, len);
	}

	// PX 20 30 ff8800
	if (line[0] != 'P' || line[1] != 'X') {
		if (line[0])
			printf("invalid header: %d %d\n", line[0], line[1]);

		return false;
	}

	int state = 0;
	int tx = 0, ty = 0;
	int r = 0, g = 0, b = 0;

	char *p = line;
	while(*p) {
		if (state == 0) {
			if (*p == 'P') {
				p++;
				state++;
			}
			else {
				break;
			}
		}
		else if (state == 1) {
			if (*p == 'X') {
				p++;
				state++;
			}
			else {
				break;
			}
		}
		else if (state == 2) {
			if (*p != ' ')
				state++;
			else
				p++;
		}
		else if (state == 3) {
			if (*p >= '0' and *p <= '9') {
				tx *= 10;
				tx += *p - '0';
				p++;
			}
			else {
				state++;
			}
		}
		else if (state == 4) {
			if (*p != ' ')
				state++;
			else
				p++;
		}
		else if (state == 5) {
			if (*p >= '0' and *p <= '9') {
				ty *= 10;
				ty += *p - '0';
				p++;
			}
			else {
				state++;
			}
		}
		else if (state == 6) {
			if (*p != ' ')
				state++;
			else
				p++;
		}
		else if (state == 7) {
			int off = ty * pp -> width * 3 + tx * 3;

			if (*p == '\n') {
				// requesting the current pixel value

				int r = pp -> fb[off + 0], g = pp -> fb[off + 1], b = pp -> fb[off + 2];

				char buffer[64];
				int len = snprintf(buffer, sizeof buffer, "PX %d %d %02x%02x%02x\n", tx, ty, r, g, b);

				return WRITE(pp -> fd, buffer, len);
			}

			r = fromhex(*p) << 4;
			p++;
			// from here it should've been state 8...12 :-)
			r |= fromhex(*p);
			p++;

			g = fromhex(*p) << 4;
			p++;
			g |= fromhex(*p);
			p++;

			b = fromhex(*p) << 4;
			p++;
			b |= fromhex(*p);
			p++;

			pp -> fb[off + 0] = r;
			pp -> fb[off + 1] = g;
			pp -> fb[off + 2] = b;

			break;
		}
	}

	return true;
}

void * tcp_receive_thread(void *p)
{
	pf_s_parameter_t *pp = (pf_s_parameter_t *)p;

	set_thread_name("src_pfCtcp");

	log(pp -> id, LL_INFO, "source pixelflood tcp handle thread started (%d)", pp -> fd);

	char buffer[4096];
	int o = 0;

	struct pollfd fds[] { { pp -> fd, POLLIN, 0 } };

	for(;!*pp->local_stop_flag;) {
		int rc = poll(fds, 1, 500);
		if (rc == 0)
			continue;

		if (o == sizeof buffer) {
			printf("buffer overflow\n");
			break;
		}

		int n = read(pp -> fd, &buffer[o], sizeof buffer - 1 - o);
		if (n <= 0) {
			// printf("fd %d read error (%d): %d (%d)\n", pp -> fd, n, errno, o);
			break;
		}

		buffer[o + n] = 0x00;
		// printf("received: |%s|\n", &buffer[o]);

		int old_o = o;
		o += n;

		// printf("before: |%s|\n", buffer);

		for(;;) {
			char *lf = strchr(&buffer[old_o], '\n');
			if (!lf)
				break;

			*lf = 0x00;

			// printf("command: %s\n", buffer);
			if (!handle_command(pp, buffer)) {
				printf("invalid PX\n");
				goto fail;
			}

			int offset_lf = lf - buffer;
			int bytes_left = o - (offset_lf + 1);

			if (bytes_left == 0) {
				o = 0;
				break;
			}

			if (bytes_left < 0) {
				printf("%d bytes left\n", bytes_left);
				goto fail;
			}

			memmove(buffer, lf + 1, bytes_left);
			o = bytes_left;
			buffer[o] = 0x00;
		}
	}

fail:
	close(pp -> fd);

	log(pp -> id, LL_INFO, "source pixelflood tcp handle thread terminating (%d)", pp -> fd);

	delete pp;

	return nullptr;
}

void * tcp_connection_thread(void *p)
{
	pf_s_parameter_t *pp = (pf_s_parameter_t *)p;

	set_thread_name("src_pfStcp");

	log(pp -> id, LL_INFO, "source pixelflood tcp listen thread started (port %d)", pp -> listen_port);

	int sfd = start_listen("0.0.0.0", pp -> listen_port, SOMAXCONN);

	struct pollfd fds[] { { sfd, POLLIN, 0 } };

	std::vector<pthread_t> handles;

	for(;!*pp->local_stop_flag;) {
		for(size_t i=0; i<handles.size();) {
			if (pthread_tryjoin_np(handles.at(i), NULL) == 0)
				handles.erase(handles.begin() + i);
			else
				i++;
		}

		if (poll(fds, 1, 500) == 0)
			continue;

		int cfd = accept(sfd, NULL, 0);
		if (cfd == -1)
			continue;

		log(pp -> id, LL_INFO, "source pixelflood tcp listen thread connection on fd %d", cfd);

		pf_s_parameter_t *ct = new pf_s_parameter_t;

		ct -> id = pp -> id;
		ct -> fd = cfd;
		ct -> fb = pp -> fb;
		ct -> width = pp -> width;
		ct -> height = pp -> height;
		ct -> local_stop_flag = pp -> local_stop_flag;
		ct -> listen_port = -1; // not applicable for thread

		pthread_t th;
		int rc = -1;
		if ((rc = pthread_create(&th, nullptr, tcp_receive_thread, ct)) != 0) {
			errno = rc;
			error_exit(true, "pthread_create failed (tcp pixelflood client thread)");
		}

		handles.push_back(th);
	}

	for(auto th : handles)
		pthread_join(th, nullptr);

	close(sfd);

	delete pp;

	log(pp -> id, LL_INFO, "source pixelflood tcp listen thread terminating");

	return nullptr;
}

source_pixelflood::source_pixelflood(const std::string & id, const std::string & descr, const std::string & exec_failure, const int listen_port, const int pixel_size_in, const pixelflood_protocol_t pp, const double max_fps, const int w_in, const int h_in, const int loglevel, std::vector<filter *> *const filters) : source(id, descr, exec_failure, max_fps, w_in * pixel_size_in, h_in * pixel_size_in, loglevel, filters), listen_port(listen_port), pixel_size(pixel_size_in), pp(pp)
{
	int size = width * height * 3;

	frame_buffer = (uint8_t *)allocate_aligned(size);
	memset(frame_buffer, 0x11, size);
}

source_pixelflood::~source_pixelflood()
{
	stop();

	free(frame_buffer);
}

void source_pixelflood::operator()()
{
	log(id, LL_INFO, "source pixelflood thread started");

	set_thread_name("src_pixelflood");

	if (pp == PP_TCP_TXT) {
		pf_s_parameter_t *ct = new pf_s_parameter_t;

		ct -> id = id;
		ct -> fd = -1; // not applicable here
		ct -> fb = frame_buffer;
		ct -> width = width;
		ct -> height = height;
		ct -> local_stop_flag = &local_stop_flag;
		ct -> listen_port = listen_port;

		int rc = -1;
		if ((rc = pthread_create(&th, NULL, tcp_connection_thread, ct)) != 0) {
			errno = rc;
			error_exit(true, "pthread_create failed (tcp pixelflood server thread)");
		}
	}

	const uint64_t interval = max_fps > 0.0 ? 1.0 / max_fps * 1000.0 * 1000.0 : 0;

	for(;!local_stop_flag;)
	{
		uint64_t start_ts = get_us();

		if (work_required() && !is_paused()) {
			if (pixel_size == 1) {
				set_frame(E_RGB, frame_buffer, width * height * 3);
			}
			else {
				uint8_t *temp_fb = (uint8_t *)allocate_aligned(width * height * 3 * pixel_size * pixel_size);

				for(int y=0; y<height; y++) {
					for(int x=0; x<width; x++) {
						int o_in = y * width * 3 + x * 3;

						uint8_t r = frame_buffer[o_in + 0];
						uint8_t g = frame_buffer[o_in + 1];
						uint8_t b = frame_buffer[o_in + 2];

						for(int yy=0; yy<pixel_size; yy++) {
							int o_out = y * width * 3 * pixel_size + yy * width * 3 + x * 3 * pixel_size;

							for(int xx=0; xx<pixel_size; x++) {
								temp_fb[o_out + xx * 3 + 0] = r;
								temp_fb[o_out + xx * 3 + 1] = g;
								temp_fb[o_out + xx * 3 + 2] = b;
							}
						}
					}
				}

				set_frame(E_RGB, temp_fb, width * height * 3 * pixel_size * pixel_size);

				free(temp_fb);
			}
		}

		uint64_t end_ts = get_us();
		int64_t left = interval - (end_ts - start_ts);

		if (interval > 0 && left > 0)
			myusleep(left);
	}

	log(id, LL_INFO, "source pixelflood thread terminating");
}
